"""
Used to manage Trello connections
"""
import requests
from cabinet import Cabinet


class TrelloManager:
    """
    A class for managing Trello boards and lists.
    """

    def __init__(self, board_name: str = None):
        self.cab = Cabinet()
        self.api_key = self.cab.get("keys", "trello")
        self.api_secret = self.cab.get("trello", "secret")
        self.api_token = self.cab.get("trello", "token")
        self.board_name = board_name or self.cab.get("trello", "board")

        if self.board_name is None:
            boards = self.list_all_boards()
            self.board_name = boards[int(
                input("\nWhich board would you like to use?\n")) - 1]['name']

    def obtain_token(self):
        """
        Obtains the token for Trello.
        """
        if self.api_token is None:
            auth_url = (
                f"https://trello.com/1/authorize?"
                f"key={self.api_key}&name=RemindMail&scope=read,write"
                "&expiration=never&response_type=token"
            )
            print(
                f"Please go to the following URL "
                f"and grant access to your Trello account:\n{auth_url}"
            )

            self.api_token = input("Enter the generated token: ")
            self.cab.put("trello", "token", self.api_token)

    def get_board_id(self, board_name):
        """
        Gets the board ID for the specified board name.

        Args:
            board_name (str): The name of the Trello board.

        Returns:
            str: The ID of the Trello board if found, None otherwise.
        """

        url = (f"https://api.trello.com/1/members/me/boards?"
               f"key={self.api_key}&token={self.api_token}")
        response = requests.get(url, timeout=10)
        response.raise_for_status()

        board_id = None
        for board in response.json():
            if board['name'] == board_name:
                board_id = board['id']
                break

        return board_id

    def get_lists_on_board(self, board_id):
        """
        Gets the lists on a Trello board.

        Args:
            board_id (str): The ID of the Trello board.

        Returns:
            list: A list of dictionaries representing the lists on the board.
        """
        url = (f"https://api.trello.com/1/boards/"
               f"{board_id}/lists?key={self.api_key}&token={self.api_token}")
        response = requests.get(url, timeout=10)
        response.raise_for_status()

        return response.json()

    def add_item_to_list(self, list_id, item_name):
        """
        Adds an item to a list.
        """
        url = "https://api.trello.com/1/cards"
        params = {
            'key': self.api_key,
            'token': self.api_token,
            'name': item_name,
            'desc': 'Added from RemindMail',
            'pos': 'top',
            'idList': list_id
        }
        response = requests.post(url, params=params, timeout=10)
        response.raise_for_status()

    def show_lists(self, is_quiet=False):
        """
        Shows a numbered list of lists in the Trello board.

        Args:
            is_quiet (bool): Whether to print the lists or not.

        Returns:
            list: A list of dictionaries representing the lists on the Trello board.
        """
        board_id = self.get_board_id(self.board_name)
        lists = None

        if board_id:
            lists = self.get_lists_on_board(board_id)
            if not is_quiet:
                for index, list_info in enumerate(lists, start=1):
                    list_name = list_info['name']
                    print(f"{index}. {list_name}")
        else:
            print(f"Board '{self.board_name}' not found.")

        return lists

    def add_item(self, list_index=None, item_name=None):
        """
        Adds an item to a Trello list.

        Args:
            list_id (str): The ID of the Trello list.
            item_name (str): The name of the item to add.
        """
        board_id = self.get_board_id(self.board_name)
        if board_id:
            lists = self.get_lists_on_board(board_id)
            if 0 <= list_index < len(lists):
                list_id = lists[list_index]['id']
                self.add_item_to_list(list_id, item_name)
                print(
                    f"\n'{item_name}' added to {lists[list_index]['name']}.")
            else:
                print("Invalid list index.")
        else:
            print(f"Board '{self.board_name}' not found.")

    def show_items(self, list_index):
        """
        Shows the items in the selected list.
        """
        board_id = self.get_board_id(self.board_name)
        if board_id:
            lists = self.get_lists_on_board(board_id)
            if 0 <= list_index < len(lists):
                list_id = lists[list_index]['id']
                url = (f"https://api.trello.com/1/lists/{list_id}/"
                       f"cards?key={self.api_key}&token={self.api_token}")
                response = requests.get(url, timeout=10)
                response.raise_for_status()
                items = response.json()
                for item in items:
                    item_name = item['name']
                    print(item_name)
            else:
                print("Invalid list index.")
        else:
            print(f"Board '{self.board_name}' not found.")

    def list_all_boards(self):
        """
        Lists all available Trello boards for the authorized user.

        Returns:
            list: A list of dictionaries representing the available boards.
        """
        url = (f"https://api.trello.com/1/members/me/boards"
               f"?key={self.api_key}&token={self.api_token}")
        response = requests.get(url, timeout=10)
        response.raise_for_status()

        boards = response.json()
        for index, board in enumerate(boards, start=1):
            board_name = board['name']
            print(f"{index}. {board_name}")

        return boards

    def main(self):
        """
        Main function.
        """
        print("Direct invokation of Trello is not yet implemented.")


if __name__ == '__main__':
    trello_manager = TrelloManager()
    trello_manager.main()



"""
Contains standalone helper functions
"""

import os
import sys
import time
import datetime
import base64
from typing import List
import requests
import pytz
from cabinet import Cabinet, Mail


class RemindMailUtils:
    """
    A utility class for handling reminders and email operations.
    """

    # instance of the Cabinet class for configuration management
    cab = Cabinet()

    # file path for reminders
    path_remind_file = cab.get('path', 'remindmail', 'file')

    # for sending emails
    mail = Mail()

    # to store query trace information
    query_trace: List[str] = []

    def get_user_time(self):
        """
        A helper function to return the epoch seconds based on the user's timezone.

        Returns:
            int: The epoch seconds representing the current time in the user's timezone.
        """

        user_timezone = self.cab.get("remindmail", "timezone")
        if user_timezone is None:
            # user's timezone not saved in `cab`; get timezone based on IP
            print("Checking timezone...")
            response = requests.get('https://ipapi.co/json/', timeout=10)

            # parse the response JSON to get the user's timezone
            if response.status_code == 200:
                data = response.json()
                user_timezone = self.cab.put(
                    "remindmail", "timezone", data.get('timezone'))
                print(f"Set timezone: {user_timezone}")
            else:
                print("Could not parse timezone; using UTC")
                return datetime.datetime.utcnow()

        timezone = pytz.timezone(user_timezone)

        now = datetime.datetime.now(timezone)

        # get the UTC time
        utc_now = datetime.datetime.utcnow()

        # calculate the time difference between the specified timezone and UTC
        offset = timezone.utcoffset(utc_now).total_seconds()

        # subtract the offset from the current time
        utc_time = now + datetime.timedelta(seconds=offset)

        # convert the UTC time to the number of seconds since January 1, 1970
        unix_time = int(utc_time.timestamp())

        return unix_time

    def get_sent_today(self):
        """
        Returns the number of reminders that have been sent today,
        based on today's log file.

        "Today", for these purposes, starts at 4AM.
        """

        today = datetime.datetime.now().strftime("%Y-%m-%d")
        yesterday = (datetime.datetime.now() -
                     datetime.timedelta(days=1)).strftime("%Y-%m-%d")

        if datetime.datetime.now().hour < 4:
            today = yesterday

        today_log = self.cab.get_file_as_array(
            f"{today}/LOG_DAILY_{today}.log", self.cab.path_log)

        return sum(log.count(': Reminder - ') for log in today_log)

    def print_stats(self):
        """
        Prints RemindMail Usage Statistics
        """

        print("RemindMail Stats:\n")
        print(f"Sent Today: {self.get_sent_today()}")

    def print_reminders_file(self, param=None):
        """
        Displays the scheduled reminders in remind.py, formatted with line numbers
        Usage: remindmail ls
        Parameters:
        - param: string; currently unused

        Passing 'help' will only return the help information for this function.
        """

        if param == "help":
            return (f"Displays the scheduled reminders in remind.py"
                    f" (in {RemindMailUtils.path_remind_file}),"
                    " formatted with line numbers\n\nUsage: remindmail ls")

        remindmd_local = f"{RemindMailUtils.path_remind_file}/remind.md"

        if RemindMailUtils.path_remind_file is not None:
            os.system(f"cat -n {remindmd_local}")
        else:
            print(f"Could not find reminder path; in "
                  f"${self.cab.path_cabinet}/settings.json, "
                  f"set path -> remindmail -> file to the absolute path of "
                  f"the directory of your remind.md file.")
            print("\n")

    def mail_reminders_for_later(self):
        """Mails a summary of reminders with [any] at the start from remind.md in {PATH_LOCAL}"""

        remindmd_file = self.cab.get_file_as_array(
            "remind.md", "notes") or []
        reminders_for_later = []
        for item in remindmd_file:
            if item.startswith("[any] "):
                reminders_for_later.append(f"<li>{item.split('[any] ')[1]}")

        mail_summary = "Just a heads up, these reminders are waiting for you!<br><br>"
        mail_summary += "".join(reminders_for_later)
        mail_summary += "<br><br>To remove these, edit <b>remind.md</b>."

        date_formatted = datetime.datetime.today().strftime('%B %d, %Y')
        RemindMailUtils().send_email(
            f"Pending Reminder Summary: {date_formatted}", mail_summary, is_quiet=False)

    def offset(self, args):
        """
        Calculates the offset for a certain date (today by default)
        """

        def months_since_epoch(epoch):
            epoch_time = time.localtime(epoch)
            return ((epoch_time.tm_year - 1970) * 12) + epoch_time.tm_mon

        if len(args) < 2:
            print(f"""Error: 'remind -o {args[0] or ''}' missing required arguments:
    remind -o '<type (day,week,month)> <target date (YYYY-MM-DD), optional> <n>'\
    \n\nRun `remind -h` for details.""")
            return

        if len(args) >= 2:
            epoch_time = int(datetime.datetime.strptime(
                args[1], "%Y-%m-%d").timestamp())
            offset_n = args[2]
        else:
            offset_n = args[-1]
            epoch_time = int(time.time())

        try:
            if not offset_n.isnumeric():
                raise IndexError

            offset_n = int(offset_n)
            token_example = "d"
            if args[0] == "month":
                token_example = "m"
                return_val = months_since_epoch(epoch_time) % offset_n
            elif args[0] == "week":
                token_example = "w"
                return_val = int(epoch_time/60/60/24/7) % offset_n
            elif args[0] == "day":
                return_val = int(epoch_time/60/60/24) % offset_n
            else:
                print(f"'{args[0]}' must be 'day', 'week', or 'month'.")
                return

            print(return_val)
            print(f"This means you can add '[{token_example}%{offset_n}+{return_val}] <task name>' \
to {RemindMailUtils.path_remind_file}/remind.md to match the selected date.")

            if offset_n == 1:
                print((f"Note: Anything % 1 is always 0. This is saying "
                       f"'every single {args[0]}'.\nOffsets are used to make sure a task will run"
                       f" for a given {args[0]}. '%1' means it will always run, so no need for an"
                       f" offset.\nPlease see the README for details, or"
                       f" run 'remindmail help offset'."))
            elif return_val == 0:
                print(("Note: The offset is 0, so a task for this date in remind.md "
                       "will be added without an offset."))

        except ValueError:
            print("Date must be YYYY-MM-DD.")
        except IndexError:
            print((f"Missing <n>, as in 'every n {args[0]}s'\n"
                   f"Usage: remindmail offset {args[0]} {args[1]} n"
                   f"\nFor help: 'remindmail help offset'"))
            return

    def edit_reminders_file(self):
        """
        Edits the remind.md file
        You must configure the path to remind.md in

        cabinet -> settings.json -> path -> edit -> remind
        """

        try:
            self.cab.edit_file("remind")
        except FileNotFoundError:
            print((f"You must configure the path to remind.md in "
                   f"{self.cab.path_cabinet}/settings.json -> "
                   f"path -> edit -> remind.\n\n"))

            resp = ''
            while resp not in ['y', 'n']:
                resp = input((f"Would you like to set this to "
                             f"{RemindMailUtils.path_remind_file}/remind.md? y/n\n\n"))
                if resp == 'y':
                    self.cab.put("path", "edit", "remind", "value",
                                 f"{RemindMailUtils.path_remind_file}/remind.md")
                    print((f"\n\nSet. Open {self.cab.path_cabinet}/settings.json"
                           f" and set path -> edit -> remind -> sync to true"
                           f" to enable cloud syncing."))
        sys.exit()

    def send_email(self, subject, body, method="Terminal", is_quiet=False):
        """A helper function to call mail.send"""

        print(f"Sending: {subject}")

        if body:
            subject = f"{subject} [See Notes]"

        body += f"<br><br>Sent via {method}"

        RemindMailUtils.mail.send(
            f"Reminder - {subject}", body or "", is_quiet=is_quiet)
        print("\nSent! Check your inbox.")

    def create_jira_issue(self, summary, description, issue_type: str = None, label: str = None):
        """
        A barebones integration with Jira.

        Requires Cabinet to be configured as described in README.

        Parameters:
            summary(str): issue title
            description(str): issue description
            issue_type(str): issue type ("story", "task", "bug", "spike", "epic")
            label(str): label
        """

        api_url_base = self.cab.get('jira', 'project-url')

        if api_url_base is None:
            self.cab.log("RemindMail - Cannot create Jira issue; cabinet",
                         "-> jira -> project-url is unset",
                         level="warn")
            return

        api_url = f"{str(api_url_base).rstrip('/')}/rest/api/2/issue"

        # Set your Jira credentials
        jira_email = self.cab.get("jira", "email")
        jira_api_token = self.cab.get("keys", "jira")

        if not jira_api_token or not jira_email:
            self.cab.log("RemindMail - cannot create Jira issue; cabinet is missing "
                         "`keys -> jira` or `jira -> email` (see README)",
                         level="warn")
            return

        if not summary:
            self.cab.log("RemindMail - cannot create Jira issue; summary cannot be blank",
                         level="warn")
            return

        # Create the authentication header
        credentials = f'{jira_email}:{jira_api_token}'
        encoded_credentials = base64.b64encode(
            credentials.encode('utf-8')).decode('utf-8')
        headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Basic {encoded_credentials}'
        }

        # Create the issue payload
        payload = {
            'fields': {
                'project': {'key': self.cab.get("jira", "project-key")},
                'summary': f"RMMJ: {summary}",
                'issuetype': {'name': issue_type.capitalize() if issue_type else ''},
                'labels': [label] if label else None,
            }
        }

        # Prompt for missing fields
        if not issue_type:
            issue_type_input = input('Please enter an issue type:\n'
                                     'story\nbug\ntask\nspike\nepic\n\n').capitalize()
            payload['fields']['issuetype']['name'] = issue_type_input

        if description is None:
            description_input = input(
                '\nPlease enter a description (or Enter for none)\n')
            payload['fields']['description'] = description_input
        else:
            payload['fields']['description'] = description

        issue_type_lower = payload['fields']['issuetype']['name'].lower()
        if issue_type_lower in ['story', 'bug']:
            if issue_type_lower == 'story':
                a_c = input('\nPlease enter an AC:\n')
                payload['fields']['customfield_10035'] = a_c

        if label is None:
            label_input = input('\nPlease enter a Label:\n')
            payload['fields']['labels'] = [label_input]

        # Send the POST request to create the issue
        response = requests.post(api_url, headers=headers,
                                 json=payload, timeout=10)

        # Check the response
        if response.status_code == 201:
            created_issue = response.json()
            issue_key = created_issue['key']
            print(f"{issue_key} has been created.")
            print(f"{api_url_base}/browse/{issue_key}")
            self.cab.log(f"\n{issue_key} has been created.", is_quiet=True)
        else:
            print('Failed to create the issue.')
            print(f'Response: {response.text}')
            self.cab.log(
                f"RemindMail unable to create Jira issue: {response}", level="warn")



"""
The data structure for Reminders
"""

class Reminder:
    """
    A class representing a reminder.

    Attributes
    ----------
    message : str
        The message for the reminder.
    next_date : date
        The date of the next occurrence of the reminder
    notes : str, optional
        Additional notes for the reminder (default is an empty string).
    is_recurring : bool, optional
        Indicates whether the reminder is recurring (default is False).

    Methods
    -------
    message
        Getter and setter method for the message attribute.
    next_date
        Getter and setter method for the next_date attribute.
    notes
        Getter and setter method for the notes attribute.
    is_recurring
        Getter and setter method for the is_recurring attribute.
    """

    def __init__(self, message, next_date, notes='', is_recurring=False):
        """
        Initializes a new Reminder instance.

        Parameters
        ----------
        message : str
            The message for the reminder.
        next_date : str
            The date of the next occurrence of the reminder in the format 'YYYY-MM-DD'.
        notes : str, optional
            Additional notes for the reminder (default is an empty string).
        is_recurring : bool, optional
            Indicates whether the reminder is recurring (default is False).
        """
        self._message = message
        self._next_date = next_date
        self._notes = notes
        self._is_recurring = is_recurring

    @property
    def message(self):
        """Getter and setter method for the message attribute."""
        return self._message

    @message.setter
    def message(self, value):
        self._message = value

    @property
    def next_date(self):
        """Getter and setter method for the next_date attribute."""
        return self._next_date

    @next_date.setter
    def next_date(self, value):
        self._next_date = value

    @property
    def notes(self):
        """Getter and setter method for the notes attribute."""
        return self._notes

    @notes.setter
    def notes(self, value):
        self._notes = value

    @property
    def is_recurring(self):
        """Getter and setter method for the is_recurring attribute."""
        return self._is_recurring

    @is_recurring.setter
    def is_recurring(self, value):
        self._is_recurring = value



"""
The main class
"""

import os
import re
import subprocess
from enum import Enum
from datetime import datetime
from datetime import timedelta
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse
from .remindmail_utils import RemindMailUtils


class GenerateType(Enum):
    """
    The options available for the `generate` function

    EMAIL will send each matching reminder
    LIST will print each matching reminder to the console
    """
    EMAIL = 'email'
    LIST = 'list'


class RemindMail:
    """
    Contains the logic for parsing, generating, and sending reminders
    """

    def log_msg(self, message, level="info", path=None, print_msg=False):
        """A wrapper for cab.log to handle the remindmail log setting"""

        path_log = RemindMailUtils.cab.get("path", "log")
        path = path or path_log
        path = f"{path}/{str(datetime.today().strftime('%Y-%m-%d'))}"

        RemindMailUtils.query_trace.append(message)

        if print_msg:
            print(message)

        RemindMailUtils.cab.log(f"remindmail: {message}", level=level, file_path=path,
                                is_quiet=level == "info")

    def generate(self, force: bool = False, dry_run: bool = False,
                 generate_date: int = None, generate_type: GenerateType = GenerateType.EMAIL):
        """
        Mails reminders from the remind.md file in {PATH_LOCAL}.
        Intended to be run from crontab (try 'remindmail generate force' to run immediately)
        """

        if generate_date is None:
            generate_date = datetime.today().replace(
                hour=0, minute=0, second=0, microsecond=0)

        today_index = generate_date.day

        day_of_month_reminders_generated = RemindMailUtils.cab.get(
            "remindmail", "day_generated")

        if day_of_month_reminders_generated == '':
            day_of_month_reminders_generated = 0

        # do not generate more than once in one day unless `remind generate force`
        if not (today_index != day_of_month_reminders_generated
                and datetime.today().hour > 3) and not force and not dry_run:
            RemindMail().log_msg(
                "Reminders have already been generated today.", level="debug")
            return

        RemindMail().log_msg("Generating reminders")

        command = ''

        current_time = RemindMailUtils().get_user_time()
        epoch_day = int(current_time/60/60/24)
        epoch_week = int(current_time/60/60/24/7)
        epoch_month = int(generate_date.month)

        remindmd_file = RemindMailUtils.cab.get_file_as_array(
            "remind.md", "notes") or []

        _remindmd_file = remindmd_file.copy()
        for index, item in enumerate(_remindmd_file):

            is_match = False
            is_command = False

            today_zero_time = generate_date.replace(
                hour=0, minute=0, second=0, microsecond=0)

            # handle commands
            if "]c" in item:
                is_command = True
                command = item.split("]c ")[1]

            # ignore anything outside of [*]
            if not re.match(r"\[(.*?)\]", item):
                continue

            _item = item

            # handle notes
            item_notes = ''
            if ':' in item:
                item_notes = ''.join(item.split(":")[1:]).strip()
                item = item.split(":")[0]

            # handle [*]n, where n is a number or "d"
            token = item.split("[")[1].split("]")[0]
            token_after = item.split("]")[1].split(" ")[0]

            if token_after == "d" or token_after == "0":
                token_after = "1"

            if token_after.isdigit():
                token_after = int(token_after)
            else:
                token_after = -1

            parsed_date = ""
            if not "%" in token and not "any" in token:
                try:
                    parsed_date = parse(
                        token, fuzzy_with_tokens=True, default=generate_date)
                except ValueError:
                    try:
                        parsed_date = parse(
                            f"{token}day", fuzzy_with_tokens=True)
                    except ValueError as error:
                        RemindMail().log_msg(
                            f"Could not parse token: {token}; {error}", level="error")

            if parsed_date and today_zero_time == parsed_date[0]:
                is_match = True

            elif "%" in token:

                if item[1:4] in ['sun', 'mon', 'tue', 'wed', 'thu', 'fri', 'sat']:
                    split_type = item[1:4]
                else:
                    split_type = item[1].lower()  # d, w, m
                split_factor = item.split("%")[1].split("]")[0]
                split_offset = 0

                # e.g. [D%4+1] for every 4 days, offset 1
                if "+" in split_factor:
                    split_offset = int(split_factor.split("+")[1])
                    split_factor = int(split_factor.split("+")[0])
                else:
                    split_factor = int(split_factor)

                is_epoch_equal_offset = (
                    split_type == 'd' and epoch_day % split_factor == split_offset,
                    split_type == 'w' and epoch_week % split_factor == split_offset,
                    split_type == 'm' and epoch_month % split_factor == split_offset,
                    split_type in ['sun', 'mon', 'tue',
                                   'wed', 'thu', 'fri', 'sat']
                    and epoch_week % split_factor == split_offset)

                today_dayw = generate_date.strftime("%a")

                split_types = ['d', 'w', 'm', 'sun', 'mon',
                               'tue', 'wed', 'thu', 'fri', 'sat']
                if split_type in split_types and any(is_epoch_equal_offset):
                    if split_type == 'd':
                        is_match = True
                    elif split_type == 'w' and today_dayw == 'Sun':
                        is_match = True
                    elif split_type == 'm' and today_index == 1:
                        is_match = True
                    elif (split_type in ['sun', 'mon', 'tue', 'wed', 'thu', 'fri', 'sat']
                          and today_dayw.lower() == split_type):
                        is_match = True

            # handle deletion and decrementing
            if is_match:
                if not is_command:
                    if generate_type == GenerateType.EMAIL and not dry_run:
                        RemindMailUtils().send_email(item.split(' ', 1)[1],
                                                     item_notes, "remind.md")
                    if dry_run and generate_type == GenerateType.EMAIL:
                        RemindMail().log_msg(
                            f"Test Mode. Not Sending: {item.split(' ', 1)[1]}", level="debug")
                    elif generate_type == GenerateType.LIST:
                        print(item)
                if token_after == 1:
                    if not dry_run:
                        RemindMail().log_msg(
                            f"Deleting item from remind.md: {item}", level="debug")
                        try:
                            remindmd_file.remove(_item)
                        except ValueError as err:
                            RemindMail().log_msg(
                                f"Could not remove from remind.md: {err}", level="error")
                    elif generate_type != GenerateType.LIST:
                        RemindMail().log_msg(f"(Test Mode. Not deleting {item})",
                                             level="debug")
                elif token_after > 1:
                    remindmd_file[index] = (item.replace(
                        f"]{token_after} ", f"]{token_after-1} "))

                if is_command:
                    if not dry_run:
                        RemindMail().log_msg(
                            f"Executing command: {command}", level="debug")
                        try:
                            # add path so things like `cabinet` calls work from crontab
                            home_dir = os.path.expanduser("~")
                            path_local_bin = os.path.join(
                                home_dir, ".local/bin")
                            os.environ['PATH'] = f"{path_local_bin}:{os.environ['PATH']}"

                            cmd_output = subprocess.check_output(
                                command, shell=True, universal_newlines=True)
                            RemindMail().log_msg(
                                f"Results: {cmd_output}", level="debug")
                        except subprocess.CalledProcessError as error:
                            RemindMail().log_msg(
                                f"Command execution failed with exit code: {error.returncode}",
                                level="error")
                            RemindMail().log_msg(
                                f"Error output: {error.output}", level="error")
                    else:
                        RemindMail().log_msg(f"Test Mode. Not executing {command})",
                                             level="debug")

        RemindMailUtils.cab.write_file("remind.md", "notes",
                                       '\n'.join(remindmd_file), is_quiet=True)

        if not dry_run:
            RemindMail().log_msg(
                f"Setting remindmail -> day_generated to {today_index}")
            RemindMailUtils.cab.put("remindmail", "day_generated", today_index)
        elif generate_type != GenerateType.LIST:
            RemindMail().log_msg(
                f"Test Mode. Not setting cabinet -> remindmail -> day_generated to {today_index}",
                level="debug"
            )

        RemindMail().log_msg("Generated tasks")

    def show_tomorrow(self):
        """
        Prints reminders from remind.md tagged with tomorow's date in YYYY-MM-DD format
        """

        tomorrow = datetime.today() + timedelta(days=1)

        # 'tomorrow' is 'today' if between 12AM and 3AM
        if datetime.now().hour < 3:
            tomorrow = datetime.today()

        tomorrow = tomorrow.replace(
            hour=0, minute=0, second=0, microsecond=0)

        RemindMail().generate(force=True, dry_run=True, generate_date=tomorrow,
                              generate_type=GenerateType.LIST)

    def parse_query(self, query=None, manual_message='', manual_date='', noconfirm=False):
        """
        Parses arguments to determine what to email or what to write to a file for later.

        Args:
            query (str): The query string to parse (default: None).
            manual_message (str): The manual message to use if query is None (default: '').
            manual_date (str): The manual date to use (default: '').
            noconfirm (bool, optional): If True, bypass the confirmation screen.
        """

        today = str(datetime.today().strftime('%Y-%m-%d'))
        weekdays = ['sunday', 'monday', 'tuesday',
                    'wednesday', 'thursday', 'friday', 'saturday']

        # helper functions
        def get_larger(string_a: str, string_b: str):
            """A helper function to return the larger string between string_a and string_b"""

            return string_a if len(string_a) > len(string_b) else string_b

        def strip_to(query):
            """A helper function to remove the portion after 'to' or 'day to'"""
            if query.startswith(' to ') or query.startswith('to ') or query.startswith('day to '):
                return ''.join(query.split('to')[1:]).strip()

            return query.strip()

        def parse_date(query: str):
            """
            Parses a date from the input query.

            Args:
                query (str): The input query containing the date information.

            Returns:
                tuple or bool: A tuple containing the parsed date/tokens, or False if parsing fails.

            Raises:
                ValueError: If an error occurs during date parsing.

            Notes:
                - The function supports parsing various date formats.
                - If the query contains the word 'tomorrow', it returns the date for the next day.
                - If a date is parsed into before 90 days ago, it is returns a date for next year.
            """

            # if before 4AM, a request for a reminder 'tomorrow' will
            # appear as a request for today
            is_after_4am = datetime.now().hour > 4

            # handle 'tomorrow'
            if 'tomorrow' in query:
                query = (datetime.now() +
                         timedelta(days=int(is_after_4am))).strftime('%F')

            # handle current day
            current_day = datetime.now().strftime('%A')
            if is_after_4am and current_day.lower() in query.lower():
                query_date = (datetime.now() +
                              timedelta(days=7)).strftime('%F')
                query = re.sub(r'\b' + re.escape(current_day) + r'\b',
                               query_date, query, flags=re.IGNORECASE)

            try:
                parsed_date = parse(query, fuzzy_with_tokens=True)

                # handle dates in the past
                days_from_now = (parsed_date[0] - datetime.today()).days
                if days_from_now < -90:
                    return (parsed_date[0] + relativedelta(years=1), parsed_date[1])

                return parsed_date

            except ValueError:
                return False

        if query is None:
            query = manual_message

        query_original = query
        RemindMailUtils.query_trace.append(query)
        query_time_token = ''  # the 'meat' between [] in remind.md
        query_notes = ''
        query_time_formatted = ''
        query_notes_formatted = ''
        is_recurring = False

        def report_query():
            print("Reporting bad query via email...")
            RemindMail().log_msg(
                f"RemindMail query reported: {query_original}", level="warn")
            RemindMailUtils().send_email(f"Bad Query: {today}", '<br>'.join(
                RemindMailUtils.query_trace).replace("\n", "<br>"), False, is_quiet=False)

        if manual_date == 'now':
            manual_date = today

        # parse for notes
        if ':' in query:
            query_notes = ''.join(query.split(":")[1:])
            query = query.split(":")[0]

        # remove filler text
        for item in ['me to ', 'to ', 'me ']:
            if item in query.lower() and len(query.split(item)[0]) < 3:
                query = re.sub(item, '', query, flags=re.IGNORECASE, count=1)

        # handle recurring reminders
        is_recurring_options = ["every [0-9]+", "every week", "every month",
                                "every day"] + [f"every {day}" for day in weekdays]

        is_recurring = any(re.findall(
            '|'.join(is_recurring_options), query, flags=re.IGNORECASE))

        if is_recurring:
            for weekday in weekdays:
                if weekday in query:
                    query_time_token = weekday[0:3].lower()
                    query_time_formatted = f"every {weekday.capitalize()}"

                    # remove 'every *day' from query
                    query = re.sub('every', '', query, flags=re.IGNORECASE)
                    query = re.sub(weekday, '', query, flags=re.IGNORECASE)

            options = [
                ('every ', 'in '),
                ('in day', 'in 1 day'),
                ('in week', 'in 1 week'),
                ('in month', 'in 1 month')
            ]
            for opt in options:
                query = re.sub(opt[0], opt[1], query, flags=re.IGNORECASE)

        # handle "in n months"
        _months = re.findall("in [0-9]+ months|in 1 month",
                             query, flags=re.IGNORECASE)

        if _months:
            _number_of_months = int(re.search(r'\d+', _months[0]).group())
            _newdate = (datetime.now().date() +
                        relativedelta(months=_number_of_months))
            _query_match = query.split(_months[0])
            query = get_larger(_query_match[0], _query_match[1])

            if is_recurring:
                query_time_token = f"M%{_number_of_months}"
                _frequency = (f"{f'{_number_of_months} ' if _number_of_months > 1 else ''}"
                              f"{'month' if _number_of_months == 1 else 'months'}")
                query_time_formatted = f"every {_frequency} starting {_newdate.strftime('%B %d')}"
            else:
                query_time_token = _newdate
                query_time_format = '%A, %B %d'

                if _newdate.year != datetime.today().year:
                    query_time_format = '%A, %B %d, %Y'
                query_time_formatted = _newdate.strftime(query_time_format)

        # handle "in n weeks"
        _weeks = re.findall("in [0-9]+ weeks|in 1 week",
                            query, flags=re.IGNORECASE)
        if _weeks:
            # reduce into "in n days"
            _number_of_weeks = int(re.search(r'\d+', _weeks[0]).group())
            query = re.sub(_weeks[0], f"in {_number_of_weeks * 7} days", query)

        # handle "in n days"
        _days = re.findall("in [0-9]+ days|in 1 day",
                           query, flags=re.IGNORECASE)
        if _days:
            _number_of_days = int(re.search(r'\d+', _days[0]).group())
            _query_match = query.split(_days[0])
            query = get_larger(_query_match[0], _query_match[1])

            _newdate = datetime.now().date() + timedelta(days=_number_of_days)
            if is_recurring:
                query_time_token = f"D%{_number_of_days}"
                query_time_formatted = (f"every {_number_of_days} days"
                                        f" starting {_newdate.strftime('%B %d')}")
            else:
                query_time_token = _newdate
                query_time_format = '%A, %B %d'

                # if in a different year, append the year
                if _newdate.year != datetime.today().year:
                    query_time_format = '%A, %B %d, %Y'
                query_time_formatted = _newdate.strftime(query_time_format)

        if " at " in query or " on " in query or " next " in query:

            # ['on sun', ... , 'on sat', 'next sun', ... , 'next sat']
            day_prefixes = [f"on {day[:3]}" for day in weekdays] + \
                [f"next {day[:3]}" for day in weekdays]
            for day in day_prefixes:
                if re.search(day, query, flags=re.IGNORECASE):

                    _query_match = re.split(day, query, flags=re.IGNORECASE)
                    query = get_larger(_query_match[0], _query_match[1])

                    query_time_token = re.sub(
                        'on|next|day', '', day, flags=re.IGNORECASE).strip()

                    query_time_formatted = {
                        'sun': 'Sunday',
                        'mon': 'Monday',
                        'tues': 'Tuesday',
                        'wednes': 'Wednesday',
                        'thurs': 'Thursday',
                        'fri': 'Friday',
                        'satur': 'Saturday'
                    }.get(query_time_token, None)

                    if query_time_formatted is None:
                        raise KeyError(
                            f"Error: '{query_time_token}' not matched to a weekday")

                    break

        # handle other dates
        parsed_date = parse_date(query)

        # handle "tomorrow"
        if not query_time_token and re.search("tomorrow", query, flags=re.IGNORECASE):
            _query_match = re.split("tomorrow", query, flags=re.IGNORECASE)
            query = strip_to(get_larger(_query_match[0], _query_match[1]))

        # handle manual time
        if manual_date:
            parsed_date = parse_date(manual_date)

        # handle specific dates (found or specified)
        if parsed_date and (not query_time_token or manual_date):
            query_time_token = parsed_date[0].strftime('%F')

            if not query_time_formatted and manual_date != today:
                query_time_format = '%A, %B %d'

                if parsed_date[0].year != datetime.today().year:
                    query_time_format = '%A, %B %d, %Y'

                query_time_formatted = parsed_date[0].strftime(
                    query_time_format)

            if manual_message:
                query = manual_message
            else:
                if len(parsed_date[1]) > 1:
                    _join_operator = ''
                    if len(parsed_date[1]) > 1:
                        _join_operator = parsed_date[1][1]
                    query = ''.join(get_larger(
                        parsed_date[1][0], _join_operator))
                    query = strip_to(''.join(query.rsplit(' on ', 1)) or query)
                elif len(parsed_date) > 1:
                    parsed_date_formatted = strip_to(''.join(parsed_date[1]))
                    if parsed_date_formatted:
                        query = parsed_date_formatted

        # confirmation
        if query_notes:
            query_notes_formatted = f"\nNotes: {query_notes.strip()}\n"

        response = ''
        query = strip_to(query.strip())

        if manual_message:
            query = manual_message

        while response not in ['y', 'n', 'r', 'l', 't', 'j', 'm']:
            options = (
                "(y)es\n"
                "(n)o\n"
                "(p)arse entire query\n"
                "(r)eport\n"
                "(l)ater\n"
                "(t)omorrow\n"
                "(j)ira\n"
                "(m)anual"
            )

            query_time_formatted = query_time_formatted or 'right now'

            if not noconfirm:
                prompt = (f"""\nYour reminder for {query_time_formatted}:"""
                          f"\n{query}\n{query_notes_formatted or ''}\nOK?\n\n{options}\n\n")
                RemindMailUtils.query_trace.append(prompt)

                response = input(prompt)
                RemindMailUtils.query_trace.append(response)
            else:
                response = 'y'

            if response == 'p':
                query_time_token = ''
                query_time_formatted = ''
                query = query_original
                print("\n------------------------------")
                RemindMailUtils.query_trace.append(
                    "\n------------------------------")

            elif response == 'l':
                query_time_token = 'any'
                is_recurring = True
                query_time_formatted = 'later'

            elif response == 'm':
                print("\n\n")
                RemindMail().manual_reminder()
                return

            elif response == 't':
                days = ["Monday", "Tuesday", "Wednesday",
                        "Thursday", "Friday", "Saturday", "Sunday"]
                weekday = days[(datetime.now() + timedelta(days=1)).weekday()]
                RemindMail().manual_reminder(query, weekday)
                return

            elif response == 'j':
                print("Creating Jira issue...")
                RemindMailUtils().create_jira_issue(summary=query, description=query_notes or '')
                return

        # send immediate reminders
        if query_time_formatted == 'right now' and response == 'y':
            RemindMailUtils().send_email(query.strip(), query_notes, False, is_quiet=True)
            return

        # scheduled reminders
        if query_time_token:
            if len(response) > 0 and not response.startswith('n'):
                if response == 'r':
                    report_query()
                else:
                    # add to remind.md file
                    query = query.strip()
                    if query_notes:
                        query = f"{query}: {query_notes}"
                    remind_md = RemindMailUtils.cab.get_file_as_array(
                        'remind.md', 'notes')
                    remind_md.append(
                        f"[{query_time_token}]{'' if is_recurring else 'd'} {query}")
                    RemindMailUtils.cab.write_file("remind.md", "notes",
                                                   '\n'.join(remind_md), is_quiet=True)
                    RemindMail().log_msg(
                        f"Scheduled \"{query.strip()}\" for {query_time_formatted}",
                        print_msg=True
                    )
            return

        if len(response) > 0:
            if response == 'r':
                report_query()
            elif not response.startswith('n'):
                # send 'right now' reminder
                RemindMailUtils().send_email(query.strip(), query_notes, is_quiet=True)

    def manual_reminder(self, manual_message='', manual_date='', noconfirm=False):
        """
        Creates a reminder with a message and a date.
        If manual_message and/or manual_date are not provided,
            prompts the user to input them.

        Args:
            manual_message (str): Optional. The reminder message.
            manual_date (str): Optional. The reminder date in a string format.

        Returns:
            None

        Example:
            manual_reminder('Buy groceries', '2023-04-03')
        """

        reminder_message = manual_message or input("What's the reminder?\n")
        reminder_date = manual_date or input(
            "\nWhen do you want to be reminded? (blank for now)\n") or "now"
        RemindMailUtils.query_trace.append(
            f"... calling parse_query({reminder_message},{reminder_date}) ")
        RemindMail().parse_query(manual_message=reminder_message,
                                 manual_date=reminder_date, noconfirm=noconfirm)



"""
The main entrypoint
"""

from pathlib import Path
import sys
import argparse
from remind.remindmail_utils import RemindMailUtils  # pylint: disable=wrong-import-position
from remind.remindmail import RemindMail  # pylint: disable=wrong-import-position
from remind.trello_manager import TrelloManager  # pylint: disable=wrong-import-position
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))


def main():
    """
    Parses command line arguments and performs the appropriate action based on the arguments passed.
    """

    help_offset = """Calculates the offset for a certain date (today by default)

		`remind -o <type (day,week,month)> <target date (YYYY-MM-DD), optional> <n>`
		(type is day, week, month)
		(n is 'every n days')

		Take the results of this function and use it to add an offset to a function.
		If you want something to happen every 3 days starting tomorrow, use:
		`remind -o day <tomorrow's date YYYY-MM-DD> 3`. See README.md for other examples."""

    parser = argparse.ArgumentParser()

    parser.add_argument('-m', '--message', nargs='?',
                        const='', help='Specify reminder message')
    parser.add_argument('-d', '--date', nargs='?', const='',
                        help='Specify reminder date')
    parser.add_argument('-n', '--notes', action='store_true',
                        help='Notes for the body of the email (or description of Jira task)')
    parser.add_argument('-ls', '--list', action='store_true',
                        help='List all reminders')
    parser.add_argument('-g', '--generate', action='store_true',
                        help='Generate reminders. By default, only generates every 12 hours.')
    parser.add_argument('--force', action='store_true',
                        help='Force reminders to be generated. Only used with -g.')
    parser.add_argument('--noconfirm', action='store_true',
                        help='Skip the confirmation before a reminder is scheduled/sent.')
    parser.add_argument('--dry-run', action='store_true',
                        help='Print generated reminders without sending them. Only used with -g.')
    parser.add_argument('--later', action='store_true',
                        help='Mail reminders for later')
    parser.add_argument('--stats', action='store_true',
                        help='Prints RemindMail usage statistics')
    parser.add_argument('-o', '--offset', nargs='?',
                        help=help_offset)
    parser.add_argument('-e', '--edit', action='store_true',
                        help='Edits remind.md through the terminal')
    parser.add_argument('--show-tomorrow', action='store_true',
                        help='Shows a list of reminders scheduled for tomorrow')
    parser.add_argument('--sent-today', action='store_true',
                        help='Prints the sum of reminders sent today (yesterday, if before 4AM)')
    parser.add_argument('manual_reminder_args', nargs='*')
    parser.add_argument('-j', '--jira', action='store_true',
                        help='Creates an issue through Jira rather than using email.')
    parser.add_argument('--type', nargs='?', const='task',
                        help=('Specify the Jira issue '
                              'type ("story", "task", "bug", "spike", "epic")'))
    parser.add_argument('--desc', '--description', nargs='?',
                        help='Jira issue description')
    parser.add_argument('--label', nargs='?',
                        help='Jira label')

    # Trello-specific Arguments
    parser.add_argument(
        "--trello-lists", "-tl", dest="trello_lists",
        action="store_true", help="Shows the lists on the board.")
    parser.add_argument(
        "--trello-items", "-ti", dest="trello_items",
        action="store_true", help="Shows items within a list.")
    parser.add_argument(
        "--trello-add", "-ta", dest="add_item",
        action="store_true", help="Add an item to the selected list.")
    parser.add_argument("--board", "-b", dest="trello_board",
                        type=str, nargs="?", help="Trello Board")
    parser.add_argument(
        "--list_index", dest="list_index", type=int, nargs="?", help="Index of the list.")
    parser.add_argument(
        "--item_name", "-i", dest="item_name", nargs="?", help="Name of the item to add.")

    args = parser.parse_args()

    # handle Trello
    if args.trello_lists:
        TrelloManager(board_name=args.trello_board).show_lists()
    elif args.add_item:
        print("Choose a list:\n")
        TrelloManager(board_name=args.trello_board).show_lists()
        list_index = int(input("\n")) - 1
        item_name = args.item_name or input(
            "\nWhat would you like to add?\n")
        TrelloManager(board_name=args.trello_board).add_item(
            list_index, item_name)
    elif args.trello_items and args.trello_items != "":
        list_index = args.list_index
        list_name = args.trello_items
        if list_index is None:
            if list_name is not None:
                for index, list_item in enumerate(TrelloManager(board_name=args.trello_board)
                                                  .show_lists(is_quiet=True)):
                    if list_item['name'].lower() == list_name:
                        list_index = index
            if list_index is None:
                print("Choose a list:\n")
                TrelloManager(board_name=args.trello_board).show_lists()
                list_index = int(input("\n")) - 1
        TrelloManager(board_name=args.trello_board).show_items(list_index)

    # handle other arguments
    else:
        if args.list:
            RemindMailUtils().print_reminders_file()
        elif args.generate:
            RemindMail().generate(force=args.force, dry_run=args.dry_run)
        elif args.later:
            RemindMailUtils().mail_reminders_for_later()
        elif args.stats:
            RemindMailUtils().print_stats()
        elif args.offset is not None:
            RemindMailUtils().offset(args.offset.split(" "))
        elif args.edit:
            RemindMailUtils().edit_reminders_file()
        elif args.show_tomorrow:
            RemindMail().show_tomorrow()
        elif args.sent_today:
            print(RemindMailUtils().get_sent_today())
        elif args.jira:
            RemindMailUtils().create_jira_issue(
                args.message or ' '.join(args.manual_reminder_args),
                args.desc or None, args.type or None, args.label)
        elif args.trello:
            TrelloManager()
        elif args.manual_reminder_args:
            RemindMail().parse_query(query=' '.join(args.manual_reminder_args),
                                     noconfirm=args.noconfirm)
        else:
            RemindMail().manual_reminder(manual_message=args.message or '',
                                         manual_date=args.date or '', noconfirm=args.noconfirm)


if __name__ == '__main__':
    main()
